--- /usr/src/sys/i386/i386/genassym.c	Wed Jun  6 11:03:25 2007
+++ i386/i386/genassym.c	Wed Jun 13 16:00:16 2007
@@ -81,6 +81,7 @@
 ASSYM(P_SFLAG, offsetof(struct proc, p_sflag));
 
 ASSYM(TD_FLAGS, offsetof(struct thread, td_flags));
+ASSYM(TD_LOCK, offsetof(struct thread, td_lock));
 ASSYM(TD_PCB, offsetof(struct thread, td_pcb));
 ASSYM(TD_PROC, offsetof(struct thread, td_proc));
 ASSYM(TD_MD, offsetof(struct thread, td_md));
--- /usr/src/sys/i386/i386/swtch.s	Sun Jun 10 23:52:42 2007
+++ i386/i386/swtch.s	Thu Jun 14 16:38:47 2007
@@ -33,6 +33,7 @@
  */
 
 #include "opt_npx.h"
+#include "opt_sched.h"
 
 #include <machine/asmacros.h>
 
@@ -91,6 +92,7 @@
  * 0(%esp) = ret
  * 4(%esp) = oldtd
  * 8(%esp) = newtd
+ * 12(%esp) = newlock
  */
 ENTRY(cpu_switch)
 
@@ -145,13 +147,16 @@
 #endif
 
 	/* Save is done.  Now fire up new thread. Leave old vmspace. */
+#ifdef SCHED_SMP
+	movl	%ecx,%edi
+	movl	12(%esp),%esi			/* New td_lock */
+#endif
 	movl	8(%esp),%ecx			/* New thread */
 #ifdef INVARIANTS
 	testl	%ecx,%ecx			/* no thread? */
 	jz	badsw3				/* no, panic */
 #endif
 	movl	TD_PCB(%ecx),%edx
-	movl	PCPU(CPUID), %esi
 
 	/* switch address space */
 	movl	PCB_CR3(%edx),%eax
@@ -160,11 +165,32 @@
 #else
 	cmpl	%eax,IdlePTD			/* Kernel address space? */
 #endif
-	je	sw1
+#ifdef SCHED_SMP
+	je	sw0
 	movl	%cr3,%ebx			/* The same address space? */
 	cmpl	%ebx,%eax
+	je	sw0
+	movl	%eax,%cr3			/* new address space */
+	movl	%esi,%eax
+	movl	PCPU(CPUID), %esi
+
+	xchgl	%eax,TD_LOCK(%edi)		/* _rel memory barrier */
+	movl	$blocked_lock,%eax
+1:
+	lock
+	cmpxchgl	%eax,TD_LOCK(%ecx)	/* _acq memory barrier */
+	jne		2f
+	pause
+	jmp	1b
+2:
+#else
+	je	sw1
+	movl	%cr3,%eax			/* The same address space? */
+	cmpl	%ebx,%eax
 	je	sw1
 	movl	%eax,%cr3			/* new address space */
+	movl	PCPU(CPUID), %esi
+#endif
 
 	/* Release bit from old pmap->pm_active */
 	movl	PCPU(CURPMAP), %ebx
@@ -183,6 +209,18 @@
 #endif
 	btsl	%esi, PM_ACTIVE(%ebx)		/* set new */
 
+#ifdef SCHED_SMP
+	jmp	sw1
+sw0:
+	xchgl	%esi,TD_LOCK(%edi)		/* _rel memory barrier */
+	movl	$blocked_lock,%eax
+1:
+	lock
+	cmpxchgl	%eax,TD_LOCK(%ecx)	/* _acq memory barrier */
+	jne		sw1
+	pause
+	jmp	1b
+#endif
 sw1:
 	/*
 	 * At this point, we've switched address spaces and are ready
